@@ -17,7 +17,7 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  *
- * $Id: pwent_mapper.c 496 2011-06-10 11:52:48Z ludovic.rousseau $
+ * $Id$
  */
 
 #define __PWENT_MAPPER_C_
@@ -167,7 +167,7 @@ mapper_module * pwent_mapper_module_init(scconf_block *blk,const char *mapper_na
 		debug= scconf_get_bool(blk,"debug",0);
 	ignorecase= scconf_get_bool(blk,"ignorecase",ignorecase);
 	} else {
-		DBG1("No block declarartion for mapper '%'",mapper_name);
+		DBG1("No block declarartion for mapper '%s'",mapper_name);
 	}
 	set_debug_level(debug);
 	pt = init_mapper_st(blk,mapper_name);
