@@ -13,7 +13,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
- * $Id: pkcs11_listcerts.c 441 2010-08-12 21:45:03Z ludovic.rousseau $
+ * $Id$
  */
 
 #ifdef HAVE_CONFIG_H
@@ -134,7 +134,7 @@ int main(int argc, const char **argv) {
     rv = verify_certificate(cert,&configuration->policy);
     if (rv < 0) {
         printf("verify_certificate() process error: %s\n", get_error());
-        goto auth_failed;
+        continue; /* try next certificate */
     } else if (rv != 1) {
         printf("verify_certificate() failed: %s\n", get_error());
         continue; /* try next certificate */
@@ -159,10 +159,4 @@ int main(int argc, const char **argv) {
 
   DBG("Process completed");
   return 0;
-
-auth_failed:
-  close_pkcs11_session(ph);
-  release_pkcs11_module(ph);
-  return 5;
-
 }
